#pragma once

#include "Singleton.h"
#include "MultiThreadHasher.h"
#include "TimerManager.h"
#include "DatabaseManager.h"
#include "HashManagerListener.h"

STANDARD_EXCEPTION(HashException);

class HashManager : public Singleton<HashManager>, public Speaker<HashManagerListener>, private TimerManagerListener, public MultiThreadHasher {
public:
        /** We don't keep leaves for blocks smaller than this... */
        static const int64_t MIN_BLOCK_SIZE = 64*1024;

        // Check if the TTH tree associated with the filename is current.
        tthroot_t checkTTH(const string& aFileName, int64_t aSize, uint32_t aTimeStamp);
        // start background hashing or read TTH from NTFS stream
        void hashFile(const string& aFileName, int64_t aSize, Flags flags = 0, DbFileObjPtr file = NULL);
        // start hashing from GUI, etc
        void hashFile(const string& aFileName, Flags flags = 0);

        // fire TTHDone event for all queued done hashes
        void flushHasher();

        bool getTTH(const string& aFileName, TTHValue& dst) throw(); // [gl] nothrow version
        bool getTree(const TTHValue& root, TigerTree& tt);
        void addTree(const TigerTree& tree);

        string getFilePath(const TTHValue& tth); // [gl]-s
        size_t getBlockSize(const TTHValue& root, int64_t* fileSize = NULL);

        void startup();
        void shutdown();

        class StreamStore { // greylink dc++: work with ntfs stream
        public:
                StreamStore(const string& aFileName) : fileName(aFileName) { }
                bool loadTree(TigerTree& tt, int64_t aFileSize = -1);
                bool saveTree(const TigerTree& tt);
        private:
                struct TTHStreamHeader {
                        uint32_t magic;
                        uint32_t checksum;      // xor of other TTHStreamHeader DWORDs
                        uint64_t fileSize;
                        uint64_t timeStamp;
                        uint64_t blockSize;
                        TTHValue root;
                };
                static const uint64_t minStreamedFileSize = 16*1048576;
                static const uint32_t MAGIC = '++lg';
                static const string streamName;
                string fileName;

                void setCheckSum(TTHStreamHeader& h);
                bool validateCheckSum(const TTHStreamHeader& h);
        };

private:
        HasherFileTask::List finished;
        CriticalSection cs;

        void hashDone(HasherFileTask::Ptr& fileTask, bool fromStream);

        // `MultiThreadHasher` override
        virtual void hashDone(HasherFileTask::Ptr& fileTask) { hashDone(fileTask, false); }

        // TimerManagerListener
        void on(Second, tick_t) throw() {}
        void on(Minute, tick_t) throw();
};
